/* This file is part of Wyslij-po
   Copyright (C) 2007, 2008, 2009 Sergey Poznyakoff

   Wyslij-po is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   Wyslij-po is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Wyslij-po.  If not, see <http://www.gnu.org/licenses/>. */

#include "wyslij-po.h"
#include <mailutils/libargp.h>

const char *program_version = "wyslij-po (" PACKAGE_STRING ")";
static char doc[] = N_("wyslij-po -- submit PO files to the Translation Project");
static char args_doc[] = N_("FILE ...");
extern char *program_invocation_short_name;

int dry_run;                    /* Do nothing, print everything */         
int verbose;                    /* Verbosity level */
char *langtab = "~/.wyslij-po.lc"; /* Name of user-specific file with language
				      names and abbreviations */
char *compress = "gzip -c";     /* Command to be used for compression */
char *address = "robot@translationproject.org"; /* E-mail of the TP robot */
char *subject_prefix = "TP-Robot "; /* Always prefix subject with this
				       string */
char *sender;                   /* Sender e-mail address */    
int set_sender_from_po;         /* If true, set sender e-mail from the PO
				   header (Last-Translator) */
char *cc;                       /* CC recipients */ 
char *bcc;                      /* BCC recipients */
char *fcc;                      /* FCC folder name */

/* What to check before submitting the PO file */
int verify_mask = VERIFY_PO_TIME | VERIFY_PACKAGE_VERSION;

#define SET_VERIFY_BIT(b) (verify_mask |= (b))
#define CLEAR_VERIFY_BIT(b) (verify_mask &= ~(b))
			   
/* Actual URL of the page of the domain %s on the TP */
char *tp_url = "http://translationproject.org/domain/${domain}.html";
/* Use this regex to guess the latest POT file name */
char *pot_regex_str =
  "The current template for this domain is <a href=\".[^\"]*/(.[^\"]*)\">";

mu_mailer_t mailer;

enum {
  LANGTAB_OPTION = 256,
  COMPRESS_OPTION,
  CC_OPTION,
  BCC_OPTION,
  FCC_OPTION,
  VERIFY_PO_TIME_OPTION,
  NO_VERIFY_PO_TIME_OPTION,
  VERIFY_VERSION_OPTION,
  NO_VERIFY_VERSION_OPTION
};

static struct argp_option options[] = {
#define GRP 10
  { NULL, 0, NULL, 0, N_("General Options:"), GRP },
  { "dry-run", 'n', NULL, 0,
    N_("do nothing, print what would have been done"), GRP+1 },
  { "verbose", 'v', NULL, 0,
    N_("produce verbose output"), GRP+1 },
  { "langtab", LANGTAB_OPTION, N_("FILE"), 0,
    N_("set the name of langtab file"), GRP+1 },
  { "compress", COMPRESS_OPTION, N_("COMMAND"), 0,
    N_("use this COMMAND for compression"), GRP+1 },
  { "gzip", 0, NULL, OPTION_ALIAS, NULL, GRP+1 },
#undef GRP
#define GRP 20
  { NULL, 0, NULL, 0, N_("Mail Address Control:"), GRP },
  { "to", 't', N_("EMAIL"), 0,
    N_("send to this email address"), GRP+1 },
  { "cc", CC_OPTION, N_("EMAIL-LIST"), 0,
    N_("set Carbon-copy recipients"), GRP+1 },
  { "bcc", BCC_OPTION, N_("EMAIL-LIST"), 0,
    N_("set Blind carbon-copy recipients"), GRP+1 },
  { "fcc", FCC_OPTION, N_("FOLDER"), 0,
    N_("direct Fcc to the given folder"), GRP+1 },
  { "from", 'F', N_("EMAIL"), OPTION_ARG_OPTIONAL,
    N_("set sender address"), GRP+1 },
#undef GRP
#define GRP 30
  { NULL, 0, NULL, 0, N_("PO File Verification Options:"), GRP },
  { "verify-po-time", VERIFY_PO_TIME_OPTION, NULL, 0,
    N_("verify PO modification time"), GRP+1 },
  { "no-verify-po-time", NO_VERIFY_PO_TIME_OPTION, NULL, 0,
    N_("disable PO modification time verification"), GRP+1 },
  { "verify-version", VERIFY_VERSION_OPTION, NULL, 0,
    N_("make sure the submitted PO file is for the latest package version"),
    GRP+1 },
  { "no-verify-version", NO_VERIFY_VERSION_OPTION, NULL, 0,
    N_("disable package version verification"), GRP+1 },
#undef GRP
#if 0
  /* This entry is to pacify `make check-docs'.  The options below
     are defined in libmailutils.
  */
  { "mailer", },
  { "help", },
  { "usage", },
  { "version" },
#endif
  { NULL }
};

static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
  switch (key)
    {
    case 'n':
      dry_run = 1;
      verbose++;
      break;

    case 't':
      address = arg;
      break;
      
    case 'v':
      verbose++;
      break;

    case LANGTAB_OPTION:
      langtab = arg;
      break;

    case COMPRESS_OPTION:
      compress = arg;
      break;

    case CC_OPTION:
      cc = arg;
      break;
      
    case BCC_OPTION:
      bcc = arg;
      break;
      
    case FCC_OPTION:
      fcc = arg;
      break;

    case 'F':
      if (arg == NULL)
	set_sender_from_po = 1;
      else
	sender = arg;
      break;

    case VERIFY_PO_TIME_OPTION:
      SET_VERIFY_BIT (VERIFY_PO_TIME);
      break;
      
    case NO_VERIFY_PO_TIME_OPTION:
      CLEAR_VERIFY_BIT (VERIFY_PO_TIME);
      break;
      
    case VERIFY_VERSION_OPTION:
      SET_VERIFY_BIT (VERIFY_PACKAGE_VERSION);
      break;
      
    case NO_VERIFY_VERSION_OPTION:
      CLEAR_VERIFY_BIT (VERIFY_PACKAGE_VERSION);
      break;

    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

static const char *capa[] = {
  "common",
  "mailer",
  NULL
};

static struct argp argp = {
  options,
  parse_opt,
  args_doc,
  doc,
  NULL,
  NULL,
  NULL
};



static int
_cb_verify_option (mu_debug_t debug, void *data, mu_config_value_t *arg,
		   int bit)
{
  int flag;

  if (mu_cfg_assert_value_type(arg, MU_CFG_STRING, debug))
    return 1;
  if (mu_cfg_parse_boolean (arg->v.string, &flag))
    {
      mu_cfg_format_error (debug, MU_DEBUG_ERROR, _("not a boolean"));
      return 1;
    }
  if (flag)
    SET_VERIFY_BIT (bit);
  else
    CLEAR_VERIFY_BIT (bit);
  return 0;
}

static int
_cb_verify_po_time (mu_debug_t debug, void *data, mu_config_value_t *arg)
{
  return _cb_verify_option (debug, data, arg, VERIFY_PO_TIME);
}
    
static int
_cb_verify_version (mu_debug_t debug, void *data, mu_config_value_t *arg)
{
  return _cb_verify_option (debug, data, arg, VERIFY_PACKAGE_VERSION);
}
    
struct mu_cfg_param wyslij_po_cfg_param[] = {
  { "compress", mu_cfg_string, &compress, 0, NULL,
    N_("Use <command> for compression."),
    N_("command") },
  { "langtab", mu_cfg_string, &langtab, 0, NULL,
    N_("Set the name of the langtab file."),
    N_("file") },
  { "to", mu_cfg_string, &address, 0, NULL,
    N_("Set recipient address. Multiple emails are allowed."),
    N_("email") },
  { "cc", mu_cfg_string, &cc, 0, NULL,
    N_("Set carbon copy recipients."),
    N_("emails") },
  { "bcc", mu_cfg_string, &bcc, 0, NULL,
    N_("Set blind carbon copy recipients."),
    N_("emails") },
  { "fcc", mu_cfg_string, &fcc, 0, NULL,
    N_("Save copy in <folder>."),
    N_("folder") },
  { "from", mu_cfg_string, &sender, 0, NULL,
    N_("Set sender address"),
    N_("email") },
  { "from-po", mu_cfg_bool, &set_sender_from_po, 0, NULL,
    N_("Read sender address from the PO file.") },
  { "verify-po-time", mu_cfg_callback, NULL, 0, _cb_verify_po_time,
    N_("Verify PO modification time"),
    N_("arg: bool") },
  { "verify-version", mu_cfg_callback, NULL, 0, _cb_verify_version,
    N_("Verify package version."),
    N_("arg: bool") },
  { "tp-url", mu_cfg_string, &tp_url, 0, NULL,
    N_("URL of textual domain page. "
       "Characters %s are replaced with the textual domain name. "
       "The default value is http://translationproject.org/domain/%s.html") },
  { "pot-regex", mu_cfg_string, &pot_regex_str, 0, NULL,
    N_("Regular expression to extract the latest POT file name from tp-url.") },
  { NULL }
};


static char *
concat (const char *a, const char *b, const char *c)
{
  size_t size = strlen (a) + strlen (b) + (c ? strlen (c) : 0 ) + 1;
  char *p = xmalloc (size);
  strcpy (p, a);
  strcat (p, b);
  if (c)
    strcat (p, c);
  return p;
}

void
create_body (mu_message_t *pmsg, const char *name, const char *gzname)
{
  mu_message_t msg;
  int status;
  mu_body_t body;
  mu_header_t hdr;
  mu_stream_t bstream; /* Body stream */
  mu_stream_t cstream; /* Compressor stream */
  mu_stream_t estream; /* Encoder stream */
  char *cmdline;
  char buffer[512]; /* FIXME: make it dynamical? */
  size_t size;
  char *s;
  
  mu_message_create (pmsg, NULL);
  msg = *pmsg;

  mu_message_get_header (msg, &hdr);
  /* mu_headder_append (hdr, MU_HEADER_CONTENT_DESCRIPTION*/
  mu_header_append (hdr, MU_HEADER_CONTENT_TRANSFER_ENCODING, "base64");
  s = concat ("application/x-gzip; name=\"", gzname, "\"");
  mu_header_append (hdr, MU_HEADER_CONTENT_TYPE, s);
  free (s);

  mu_message_get_body (msg, &body);
  mu_body_get_stream (body, &bstream);

  cmdline = concat (compress, " ", name);
  if (verbose > 1) 
    printf (_("executing %s...\n"), cmdline);
  
  status = mu_prog_stream_create (&cstream, cmdline, MU_STREAM_READ);
  if (status)
    {
      mu_error (_("%s: cannot create stream for `%s': %s"),
		name, cmdline, mu_strerror (status));
      exit (1);
    }
  
  status = mu_filter_create (&estream, cstream, "base64", MU_FILTER_ENCODE,
			     MU_STREAM_READ);
  if (status)
    {
      mu_error (_("%s: cannot create base64 encoder stream: %s"),
		name, mu_strerror (status));
      exit (1);
    }

  status = mu_stream_open (estream);
  if (status)
    {
      mu_error (_("%s: cannot open base64 encoder stream: %s"),
		name, mu_strerror (status));
      exit (1);
    }

  mu_stream_seek (estream, 0, SEEK_SET);
  mu_stream_seek (bstream, 0, SEEK_SET);
  while (mu_stream_sequential_read (estream, buffer, sizeof (buffer),
				    &size) == 0
	 && size > 0)
    {
      mu_stream_sequential_write (bstream, buffer, size);
    }

  mu_stream_sequential_write (bstream, "\n", 1);
  mu_stream_close (estream);
}

void
wyslij_po (const char *name)
{
  mu_mime_t mime;
  mu_message_t msg;
  mu_header_t hdr;
  mu_stream_t stream;
  char *po_sender;
  char *poname = restore_file_name (name, &po_sender);
  char *s;
  char buf[512];
  size_t n;
  
  if (verbose)
    printf ("%s: %s\n", name, poname);
  
  mu_mime_create (&mime, NULL, 0);

  s = concat (poname, ".gz", NULL);
  create_body (&msg, name, s);
  free (s);
  mu_mime_add_part (mime, msg);
  mu_message_unref (msg);

  mu_mime_get_message (mime, &msg);
  mu_message_get_header (msg, &hdr);
  if (set_sender_from_po)
    mu_header_append (hdr, MU_HEADER_FROM, po_sender);
  else
    mu_header_append (hdr, MU_HEADER_FROM, sender);
  mu_header_append (hdr, MU_HEADER_TO, address);
  s = concat (subject_prefix, poname, NULL);
  if (cc)
    mu_header_append (hdr, MU_HEADER_CC, cc);
  if (bcc)
    mu_header_append (hdr, MU_HEADER_BCC, bcc);
  if (fcc)
    mu_header_append (hdr, MU_HEADER_FCC, fcc);
  mu_header_append (hdr, MU_HEADER_SUBJECT, s);
  free (s);
  mu_header_append (hdr, "X-Mailer", PACKAGE_STRING);

  if (verbose)
    {
      if (verbose > 3)
	{
	  printf ("\n%s\n", _("Message:"));
	  mu_message_get_stream (msg, &stream);
	}
      else
	{
	  printf ("\n%s\n", _("Message headers:"));
	  mu_header_get_stream (hdr, &stream);
	}

      mu_stream_seek (stream, 0, SEEK_SET);
      while (mu_stream_sequential_read (stream, buf, sizeof buf, &n) == 0
	     && n > 0)
	{
	  fwrite (buf, 1, n, stdout);
	}
    }

  if (dry_run)
    printf (_("NOT submitting\n"));
  else
    {
      int status;
      if (verbose)
	printf (_("Submitting %s\n"), name);
      if ((status = mu_mailer_send_message (mailer, msg, NULL, NULL)))
	{
	  mu_error (_("%s: Sending message failed: %s"),
		    name, mu_strerror (status));
	  exit (1);
	}
    }

  mu_mime_destroy (&mime);
  free (poname);
  free (po_sender);
}

static void
parse_langtabs ()
{
  char *name;

  if (access (DEFAULT_LANGTAB_FILE, R_OK) == 0)
    parse_lang_datafile (DEFAULT_LANGTAB_FILE);

  name = mu_tilde_expansion (langtab, "/", NULL);
  
  if (access (name, R_OK) == 0)
    parse_lang_datafile (name);
}

static int
_error_printer (void *unused, mu_log_level_t level, const char *msg)
{
  fprintf (stderr, "%s: ", program_invocation_short_name);
  if (level == MU_DEBUG_ERROR)
    fprintf (stderr, "%s: ", _("DEBUG"));
  fprintf (stderr, "%s", msg);
  return 0;
}

void
open_mailer ()
{
  int status;
  int mailer_flags = 0;
  mu_property_t property = NULL;

  if (dry_run)
    return;
  
  if ((status = mu_mailer_create (&mailer, NULL)))
    {
      const char *url = NULL;
      mu_mailer_get_url_default (&url);
      mu_error (_("Creating mailer `%s' failed: %s"),
		url, mu_strerror (status));
      exit (1);
    }
  if (verbose > 2)
    {
      mu_debug_t debug;
      mu_mailer_get_debug (mailer, &debug);
      mu_debug_set_level (debug, MU_DEBUG_TRACE | MU_DEBUG_PROT);
      if (verbose > 3)
	mailer_flags = MAILER_FLAG_DEBUG_DATA;
    }
  
  mu_mailer_get_property (mailer, &property);
  mu_property_set_value (property, "READ_RECIPIENTS", "true", 1);
  
  if ((status = mu_mailer_open (mailer, mailer_flags)))
    {
      const char *url = NULL;
      mu_mailer_get_url_default (&url);
      mu_error (_("Opening mailer `%s' failed: %s"),
		url, mu_strerror (status));
      exit (1);
    }
}

void
close_mailer ()
{
  int status;
  
  if (!mailer)
    return;
  
  if ((status = mu_mailer_close (mailer)))
    {
      mu_error (_("Closing mailer failed: %s"),
		mu_strerror (status));
      exit (1);
    }

  mu_mailer_destroy (&mailer);
}

const char version_etc_copyright[] =
  /* Do *not* mark this string for translation.  %s is a copyright
     symbol suitable for this locale, and %d is the copyright
     year.  */
  "Copyright %s 2009 Sergey Poznyakoff";

void
version_out (FILE *stream, struct argp_state *state)
{
	fprintf (stream, "%s (%s) %s\n", program_invocation_short_name,
		 PACKAGE, PACKAGE_VERSION);
	/* TRANSLATORS: Translate "(C)" to the copyright symbol
	   (C-in-a-circle), if this symbol is available in the user's
	   locale.  Otherwise, do not translate "(C)"; leave it as-is.  */
	fprintf (stream, version_etc_copyright, _("(C)"));

	fputs (_("\
\n\
License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n\
This is free software: you are free to change and redistribute it.\n\
There is NO WARRANTY, to the extent permitted by law.\n\
\n\
"),
	       stream);
	
	/* TRANSLATORS: %s denotes an author name.  */
	fprintf (stream, _("Written by %s.\n"), "Sergey Poznyakoff");
}

int
main (int argc, char **argv)
{
  int index;
  mu_debug_t debug;
  
  mu_register_all_mailer_formats ();
  if (!program_invocation_short_name)
    program_invocation_short_name = argv[0];
  argp_program_version_hook = version_out;
  mu_argp_init (program_version, "<" PACKAGE_BUGREPORT ">");
  if (mu_app_init (&argp, capa, wyslij_po_cfg_param, argc, argv, 0, &index,
		   NULL))
    exit (1);

  mu_diag_get_debug (&debug);
  mu_debug_set_print (debug, _error_printer, NULL);
  
  if (index == argc)
    {
      mu_error (_("Not enough arguments"));
      exit (1);
    }

  if (!sender && !set_sender_from_po)
    sender = mu_get_user_email (NULL);
  
  parse_langtabs ();
  compile_pot_regex_str ();
  
  open_mailer ();
  
  for (; index < argc; index++)
    wyslij_po (argv[index]);

  close_mailer ();
  
  exit (0);
}
